from rest_framework import serializers
from rest_framework.serializers import ModelSerializer

from datapunt_api.rest import HALSerializer
from datapunt_api.serializers import (DataSetSerializerMixin, DisplayField,
                                      MultipleGeometryField,
                                      RelatedSummaryField,
                                      SelfLinkSerializerMixin)
from tests import models


class WeatherStationSerializer(HALSerializer):
    class Meta:
        model = models.WeatherStation
        fields = '__all__'


class DatasetSerializer(DataSetSerializerMixin, ModelSerializer):
    dataset = 'test_dataset'

    class Meta:
        model = models.SimpleModel
        fields = '__all__'


class SelfLinksSerializer(SelfLinkSerializerMixin, ModelSerializer):
    _links = serializers.SerializerMethodField()

    class Meta:
        model = models.SimpleModel
        fields = '__all__'


class WeatherDetailStationSerializer(HALSerializer):

    detailed = serializers.SerializerMethodField()

    class Meta:
        model = models.WeatherStation
        fields = [
            '_links',
            'number',
            'detailed'
        ]

    def get_detailed(self, obj):
        return 'I am detailed'


class TemperatureRecordSerializer(HALSerializer):
    class Meta:
        model = models.TemperatureRecord
        fields = '__all__'


class DisplayFieldSerializer(ModelSerializer):
    _display = DisplayField()

    class Meta:  # noqa
        model = models.WeatherStation
        fields = '__all__'


class ThingSerializer(ModelSerializer):
    class Meta:
        model = models.Thing
        fields = '__all__'


class PersonSerializer(ModelSerializer):
    things = RelatedSummaryField()

    class Meta:
        model = models.Person
        fields = '__all__'


class LocationSerializer(ModelSerializer):
    geometrie = MultipleGeometryField()

    class Meta:
        model = models.Location
        fields = '__all__'


class DetailedPersonSerializer(ModelSerializer):
    things = RelatedSummaryField()
    detailed = serializers.SerializerMethodField()

    class Meta:
        model = models.Person
        fields = '__all__'

    def get_detailed(self, obj: models.Person) -> str:
        return "Yes, detailed isn't it?"



from collections import OrderedDict
from typing import Any, TypeVar

from django.db.models import Model, QuerySet
from rest_framework import pagination, response
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.utils.urls import replace_query_param
from rest_framework.views import APIView

_MT = TypeVar("_MT", bound=Model)


class HALPagination(pagination.PageNumberPagination):
    """Implement HAL-JSON style pagination.

    used on most rest Datapunt APIs.
    """

    page_size_query_param: str = 'page_size'

    def get_paginated_response(self, data: Any) -> Response:
        assert self.request is not None

        self_link = self.request.build_absolute_uri()
        if self_link.endswith(".api"):
            self_link = self_link[:-4]

        assert self.page is not None

        if self.page.has_next():
            next_link = replace_query_param(
                self_link, self.page_query_param, self.page.next_page_number())
        else:
            next_link = None

        if self.page.has_previous():
            prev_link = replace_query_param(
                self_link, self.page_query_param,
                self.page.previous_page_number())
        else:
            prev_link = None

        return response.Response(OrderedDict([
            ('_links', OrderedDict([
                ('self', dict(href=self_link)),
                ('next', dict(href=next_link)),
                ('previous', dict(href=prev_link)),
            ])),
            ('count', self.page.paginator.count),
            ('results', data)
        ]))


class HALCursorPagination(pagination.CursorPagination):
    """Implement HAL-JSON Cursor style pagination.

    standard for large datasets Datapunt APIs.
    """
    page_size_query_param: str = 'page_size'
    count_table: bool = True
    count: int = 0

    def paginate_queryset(
            self,
            queryset: QuerySet[_MT],
            request: Request,
            view: APIView | None = None
    ) -> list[_MT] | None:
        if self.count_table:
            self.count = queryset.count()

        return super(HALCursorPagination, self).paginate_queryset(queryset, request, view=view)

    def get_paginated_response(self, data: Any) -> Response:
        next_link = self.get_next_link()
        previous_link = self.get_previous_link()

        _response = OrderedDict([
            ('_links', OrderedDict([
                ('next', dict(href=next_link)),
                ('previous', dict(href=previous_link)),
            ])),
            ('count', self.count),
            ('results', data)
        ])

        if not self.count_table:
            _response.pop('count')

        return response.Response(_response)



from unittest.mock import Mock

from django.contrib.gis.geos import Point
from django.test import TestCase
from rest_framework.exceptions import ValidationError

from datapunt_api.bbox import (BBOX, determine_bbox, dist_to_deg, parse_xyr,
                               valid_bbox)


class ValidBboxTestCase(TestCase):
    def test_valid_bbox(self):
        bbox, err = valid_bbox("4.58565,52.03560,5.31360,52.10000")
        self.assertEqual(bbox, [4.58565, 52.03560, 5.31360, 52.10000])
        self.assertIsNone(err)

    def test_invalid_bbox(self):
        # Too few values
        bbox, err = valid_bbox("invalid_bbox")
        self.assertEqual(bbox, [])
        self.assertEqual(err, "wrong numer of arguments (lon, lat, lon, lat)")

        bbox, err = valid_bbox("1,2,3")
        self.assertEqual(bbox, [])
        self.assertEqual(err, "wrong numer of arguments (lon, lat, lon, lat)")

        # Too many values
        bbox, err = valid_bbox("1,2,3,4,5,6,7")
        self.assertEqual(bbox, [])
        self.assertEqual(err, "wrong numer of arguments (lon, lat, lon, lat)")

        # Values are not floats
        bbox, err = valid_bbox("a,b,c,d")
        self.assertEqual(bbox, [])
        self.assertEqual(err, "Did not receive floats")

        # Mixed values
        bbox, err = valid_bbox("1,2.0,A,B")
        self.assertEqual(bbox, [])
        self.assertEqual(err, "Did not receive floats")

    def test_valid_bbox_invalid_coordinates(self):
        # Test case with coordinates outside the bounding box
        bboxp = "3.6,52.1,5.2,52.4"
        bbox, err = valid_bbox(bboxp)
        self.assertEqual(bbox, [3.6, 52.1, 5.2, 52.4])
        self.assertIsNotNone(err)
        self.assertEqual(err, "lon not within max bbox 5.3136 > 3.6 > 4.58565")

        bboxp = "5.2,52.1,5.7,52.4"
        bbox, err = valid_bbox(bboxp)
        self.assertEqual(bbox, [5.2, 52.1, 5.7, 52.4])
        self.assertIsNotNone(err)
        self.assertEqual(err, "lon not within max bbox 5.3136 > 5.7 > 4.58565")

        bboxp = "4.9,51.1,5.2,52.4"
        bbox, err = valid_bbox(bboxp)
        self.assertEqual(bbox, [4.9, 51.1, 5.2, 52.4])
        self.assertIsNotNone(err)
        self.assertEqual(err, "lat not within max bbox 52.48769 > 51.1 > 52.0356")

        bboxp = "4.9,52.1,5.2,53.4"
        bbox, err = valid_bbox(bboxp)
        self.assertEqual(bbox, [4.9, 52.1, 5.2, 53.4])
        self.assertIsNotNone(err)
        self.assertEqual(err, "lat not within max bbox 52.48769 > 53.4 > 52.0356")

    def test_valid_bbox_invalid_coordinates_28992(self):
        # Test case with coordinates outside the bounding box
        bboxp = "93000,470000,96000,485000"
        bbox, err = valid_bbox(bboxp, srid=28992)
        self.assertEqual(bbox, [93000, 470000, 96000, 485000])
        self.assertIsNotNone(err)
        self.assertEqual(err, "lon not within max bbox 170000 > 93000.0 > 94000")

        bboxp = "95000,470000,180000,485000"
        bbox, err = valid_bbox(bboxp, srid=28992)
        self.assertEqual(bbox, [95000, 470000, 180000, 485000])
        self.assertIsNotNone(err)
        self.assertEqual(err, "lon not within max bbox 170000 > 180000.0 > 94000")

        bboxp = "95000,460000,96000,485000"
        bbox, err = valid_bbox(bboxp, srid=28992)
        self.assertEqual(bbox, [95000, 460000, 96000, 485000])
        self.assertIsNotNone(err)
        self.assertEqual(err, "lat not within max bbox 514000 > 460000.0 > 465000")

        bboxp = "95000,470000,96000,520000"
        bbox, err = valid_bbox(bboxp, srid=28992)
        self.assertEqual(bbox, [95000, 470000, 96000, 520000])
        self.assertIsNotNone(err)
        self.assertEqual(err, "lat not within max bbox 514000 > 520000.0 > 465000")


class DetermineBboxTestCase(TestCase):
    def test_default_bbox(self):
        request = Mock()
        request.query_params = {}

        bbox, err = determine_bbox(request)
        self.assertEqual(bbox, BBOX)
        self.assertIsNone(err)

    def test_valid_bbox(self):
        request = Mock()
        request.query_params = {"bbox": "4.58565,52.03560,5.31360,52.10000"}

        bbox, err = determine_bbox(request)
        self.assertEqual(bbox, [4.58565, 52.03560, 5.31360, 52.10000])
        self.assertIsNone(err)

    def test_invalid_bbox(self):
        # Too few values
        request = Mock()
        request.query_params = {"bbox": "invalid_bbox"}

        bbox, err = determine_bbox(request)
        self.assertIsNone(bbox)
        self.assertEqual(err, "wrong numer of arguments (lon, lat, lon, lat)")

        request.query_params = {"bbox": "1,2"}
        bbox, err = determine_bbox(request)
        self.assertIsNone(bbox)
        self.assertEqual(err, "wrong numer of arguments (lon, lat, lon, lat)")

        # Too many values
        request.query_params = {"bbox": "1,2,3,4,5"}
        bbox, err = determine_bbox(request)
        self.assertIsNone(bbox)
        self.assertEqual(err, "wrong numer of arguments (lon, lat, lon, lat)")

        # Values are not floats
        request.query_params = {"bbox": "a,b,c,d"}
        bbox, err = determine_bbox(request)
        self.assertIsNone(bbox)
        self.assertEqual(err, "Did not receive floats")

        # Mixed values
        request.query_params = {"bbox": "1,2.0,A,B"}
        bbox, err = determine_bbox(request)
        self.assertIsNone(bbox)
        self.assertEqual(err, "Did not receive floats")


class DistanceToDegreeTestCase(TestCase):
    def test_distance_lat_0(self):
        lat = 0.0
        distance = 1000
        expected_output = 0.00898312

        result = dist_to_deg(distance, lat)
        self.assertAlmostEqual(result, expected_output, places=8)  # Allowing for some rounding error

    def test_distance_lat_45(self):
        lat = 45.0
        distance = 500
        expected_output = 0.00526219

        result = dist_to_deg(distance, lat)
        self.assertAlmostEqual(result, expected_output, places=8)  # Allowing for some rounding error

    def test_distance_lat_minus_75(self):
        lat = -75.0
        distance = 1264
        expected_output = 0.01804019

        result = dist_to_deg(distance, lat)
        self.assertAlmostEqual(result, expected_output, places=8)  # Allowing for some rounding error

    def test_distance_lat_53_25367(self):
        lat = 52.25367
        distance = 80
        expected_output = 0.000891532

        result = dist_to_deg(distance, lat)
        self.assertAlmostEqual(result, expected_output, places=8)  # Allowing for some rounding error


class ParseXYRTestCase(TestCase):
    def test_valid_input_srid(self):
        input_string = '52.03560,4.58565,45.0'
        expected_point = Point(x=52.03560, y=4.58565, srid=4326)

        point, radius = parse_xyr(input_string)
        self.assertEqual(point, expected_point)
        self.assertAlmostEqual(radius, 0.00050055, places=8)  # Allowing for some rounding error

    def test_valid_input_srid_28992(self):
        input_string = '123456.789,987654.321,45.0'
        expected_point = Point(x=123456.789, y=987654.321, srid=28992).transform(ct=4326, clone=True)

        point, radius = parse_xyr(input_string)
        self.assertEqual(point, expected_point)
        self.assertEqual(radius, 45.0)

    def test_invalid_input_missing_radius(self):
        # Test with invalid input format: "12.34,56.78" (missing radius)
        input_string = "12.34,56.78"
        with self.assertRaises(ValidationError):
            parse_xyr(input_string)

    def test_invalid_input_type(self):
        # Test with input containing non-numeric values: "12.34,invalid,100"
        input_string = "12.34,invalid,100"
        with self.assertRaises(ValidationError):
            parse_xyr(input_string)

    def test_invalid_input_type_radius(self):
        # Test with invalid radius type (should be a float): "12.34,56.78,invalid"
        input_string = "12.34,56.78,invalid"
        with self.assertRaises(ValidationError):
            parse_xyr(input_string)



import json
from datetime import date

from django.contrib.gis.geos import Point
from django.test import TestCase
from rest_framework.test import APIClient

from datapunt_api.serializers import get_links
from tests.models import (Location, Person, SimpleModel, TemperatureRecord,
                          Thing, WeatherStation)
from tests.serializers import (DatasetSerializer, DisplayFieldSerializer,
                               LocationSerializer, TemperatureRecordSerializer,
                               WeatherStationSerializer)

# # fake requests
# See: https://stackoverflow.com/questions/34438290/
# from rest_framework.request import Request
# from rest_framework.test import APIRequestFactory
#
# factory = APIRequestFactory()
# request = factory.get('/')

FORMATS = [
    ('api', 'text/html; charset=utf-8'),
    ('json', 'application/json'),
    ('xml', 'application/xml; charset=utf-8'),
    ('csv', 'text/csv; charset=utf-8'),
]

BBOX = [52.03560, 4.58565,
        52.48769, 5.31360]

RD = [121357.68, 487342.57]


def get_wgs_puntje():  # noqa
    lat = BBOX[0]
    lon = BBOX[1]
    return Point(float(lat), float(lon))

def get_rd_puntje(): # noqa
    x = RD[0]
    y = RD[1]
    return Point(x, y, srid=28992)


class SerializerTest(TestCase):
    """test datapunt_api."""

    def setUp(self):
        """Create some fake data."""
        ws = WeatherStation.objects.create(
            number=260,
            centroid=get_wgs_puntje(),
            centroid_rd=get_rd_puntje()
        )

        records = [
            {'station': ws, 'date': date(1901, 1, 1), 'temperature': '10.0'},
            {'station': ws, 'date': date(1901, 2, 1), 'temperature': '11.0'},
            {'station': ws, 'date': date(1901, 3, 1), 'temperature': '20.0'},
        ]
        for rs in records:
            TemperatureRecord.objects.create(**rs)

    def valid_response(
            self, url, response,
            content_type='text/html; charset=utf-8'):
        """Help method to check common status/json."""
        self.assertEqual(
            200, response.status_code, "Wrong response code for {}".format(url)
        )

        self.assertEqual(
            f"{content_type}",
            response["Content-Type"],
            "Wrong Content-Type for {}".format(url),
        )

    def test_rd_filter(self):
        location = "%s,%s,%s" % (
            RD[0], RD[1], 6
        )
        params = {
            location: location
        }
        c = APIClient()
        url = '/tests/weatherstation/'
        response = c.get(url, params=params)
        self.valid_response(url, response, content_type='application/json')
        self.assertEqual(len(response.json()['results']), 1)

        location = "%s,%s,%s" % (
            RD[1]+100, RD[0], 6
        )

        invalid_params = {location: location}

        response = c.get(url, params=invalid_params)

        self.assertEqual(
            200, response.status_code, "Wrong response code for {}".format(url)
        )

    def test_cannot_serialize_without_request_context(self):
        qs = TemperatureRecord.objects.all()
        self.assertTrue(qs.count() == 3)

        # We need request in context to build hyperlinks for relations.
        with self.assertRaises(AssertionError):
            TemperatureRecordSerializer(qs, many=True).data

    def test_json_html(self):
        c = APIClient()
        response = c.get('/tests/')
        self.assertEqual(response.status_code, 200)

    def test_options_html(self):
        c = APIClient()
        response = c.options('/tests/')
        self.assertEqual(response.status_code, 200)
        response = c.options('/tests/weatherstation/')
        self.assertEqual(response.status_code, 200)

    def test_formats(self):

        c = APIClient()

        for fmt, encoding in FORMATS:
            url = '/tests/weatherstation/'
            response = c.get(url, {'format': fmt})
            self.valid_response(url, response, content_type=encoding)

    def test_detailed(self):
        c = APIClient()
        url = '/tests/weatherstation/'
        response = c.get(url, {'format': 'json', 'detailed': True})
        self.valid_response(url, response, 'application/json')
        data = response.json()
        self.assertEqual(data['results'][0]['detailed'], 'I am detailed')

    def test_weatherstation_endpoint(self):
        c = APIClient()
        response = c.get('/tests/weatherstation/', {'format': 'json'})
        self.assertEqual(response.status_code, 200)
        payload = json.loads(response.content)

        # check that the basics look good for list endpoint
        self.assertEqual(payload['count'], 1)
        self.assertEqual(len(payload['results']), 1)
        self.assertEqual(
            payload['_links']['self']['href'],
            'http://testserver/tests/weatherstation/?format=json'
        )

        # check that we can follow the link to the detail endpoint
        detail_page = payload['results'][0]['_links']['self']['href']

        response = c.get(detail_page, format='json')
        self.assertEqual(response.status_code, 200)
        payload = json.loads(response.content)
        self.assertEqual(
            payload['_links']['self']['href'],
            detail_page
        )
        self.assertEqual(payload['number'], 260)

    def test_hal_style_pagination(self):
        c = APIClient()
        response = c.get('/tests/weatherstation/', format='json')
        self.assertEqual(response.status_code, 200)
        payload = json.loads(response.content)

        self.assertIn('_links', payload)
        self.assertEqual(
            payload['_links']['self']['href'],
            'http://testserver/tests/weatherstation/'
        )
        self.assertEqual(payload['_links']['next']['href'], None)
        self.assertEqual(payload['_links']['previous']['href'], None)

    def test_display_field(self):
        ws = WeatherStation.objects.get(number__exact=260)
        serializer = DisplayFieldSerializer(ws)
        self.assertEqual(
            serializer.data['_display'],
            'DISPLAY FIELD CONTENT'
        )

    def test_get_links(self) -> None:
        links = get_links('weatherstation-list')
        link_self = links.get('self')

        self.assertIsNotNone(link_self)
        self.assertEqual(link_self.get('href'), '/tests/weatherstation/')

    def test_dataset_serializer(self) -> None:
        simple = SimpleModel()
        simple.name = 'Name1'
        simple.age = 13
        simple.sign = '++'

        serializer = DatasetSerializer(simple)
        dataset = serializer.data.get('dataset')
        self.assertEqual(dataset, serializer.dataset)

    def test_self_links_serializer(self) -> None:
        simple = SimpleModel()
        simple.name = 'Name1'
        simple.age = 13
        simple.sign = '++'
        simple.save()

        client = APIClient()
        response = client.get('/tests/simple/', format='json')

        body = response.json()
        results = body.get('results')
        self.assertIsNotNone(results)
        self.assertEqual(results[0].get('_links'), {
            'self': {
                'href': 'http://testserver/tests/simple/1/'
            }
        })

    def test_related_summary_field(self) -> None:
        fokje = Person()
        fokje.name = 'Fokje'
        fokje.save()

        modder = Thing()
        modder.name = 'modder'
        modder.person = fokje
        modder.save()

        client = APIClient()
        response = client.get('/tests/person/', format='json')

        body = response.json()
        results = body.get('results')
        self.assertIsNotNone(results)
        self.assertEqual(results[0], {
            'id': 1,
            'name': 'Fokje',
            'things': {
                'count': 1,
                'href': 'http://testserver/tests/thing/?person=1'
            }
        })

    def test_multiple_geometry_field(self) -> None:
        location = Location()
        location.geometrie = Point(5, 23)

        serializer = LocationSerializer(location)

        self.assertEqual(serializer.data.get('geometrie'), {
            'type': 'Point',
            'coordinates': [5.0, 23.0]
        })

    def test_detailed_serializer(self) -> None:
        fokje = Person()
        fokje.name = 'Fokje'
        fokje.save()

        modder = Thing()
        modder.name = 'modder'
        modder.person = fokje
        modder.save()

        client = APIClient()
        response = client.get('/tests/person/?detailed=True', format='json')

        body = response.json()
        results = body.get('results')
        self.assertIsNotNone(results)

        self.assertEqual(results[0], {
            'id': 1,
            'name': 'Fokje',
            'things': {
                'count': 1,
                'href': 'http://testserver/tests/thing/?person=1'
            },
            'detailed': "Yes, detailed isn't it?"
        })

    def test_view_with__DisabledHTMLFilterBackend(self):
        client = APIClient()
        response = client.get('/tests/weatherstation/')

        self.assertNotIn('<select', response.content.decode('utf-8'))

    def test_links_field_might_produce_none_url(self):
        station = WeatherStation()
        station.number = 260
        station.centroid = get_wgs_puntje()
        station.centroid_rd = get_rd_puntje()

        serializer = WeatherStationSerializer(station)

        _links = serializer.data.get('_links')
        self.assertIsNotNone(_links)

        self_links = _links.get('self')
        self.assertIsNotNone(self_links)

        self.assertIsNone(self_links.get('href'))



from typing import Any, Mapping

from rest_framework_csv.renderers import CSVRenderer


class PaginatedCSVRenderer(CSVRenderer):
    results_field: str = 'results'

    def render(
            self,
            data: Any,
            media_type: str | None = None,
            renderer_context: Mapping[str, Any] | None = {},
            writer_opts: Mapping[str, Any] | None = None,
    ) -> Any:
        if not isinstance(data, list):
            data = data.get(self.results_field, [])

        return super(PaginatedCSVRenderer, self).render(data, media_type, renderer_context, writer_opts)



"""Bounding box methods useful for Amsterdam."""
from math import cos, pi

from django.contrib.gis.geos import GEOSGeometry, Point
from rest_framework.request import Request
from rest_framework.serializers import ValidationError

# A BBOX, or "bounding box," is a rectangular area used to define a specific
# object, area, or geographic location on a map. It is defined by the
# coordinates of two opposite corners: the lower-left corner (minimum values)
# and the upper-right corner (maximum values).
#
# The BBOX coordinates are as follows:
#
# Lower-left corner (minimum values): [52.03560, 4.58565]
# Upper-right corner (maximum values): [52.48769, 5.31360]
#
# This bounding box encompasses the geographic region of Amsterdam in the
# Netherlands in the format [min_lon, min_lat, max_lon, max_lat]
BBOX = [52.03560, 4.58565, 52.48769, 5.31360]


def parse_xyr(value: str) -> tuple[GEOSGeometry, float]:
    """
    Parse x, y, radius input.

    Args:
        value: A string containing the x, y, and radius inputs, separated by commas.

    Returns:
        A tuple of a Point object and a radius, in meters.
    """
    # Split the input string into three parts: x, y, and radius.
    try:
        x_str, y_str, radius_str = value.split(',')
    except ValueError:
        raise ValidationError(
            "Locatie must be rdx,rdy,radius(m) or lat,long,radius(m)"
        )

    # Try converting the input strings to floats.
    try:
        x = float(x_str)
        y = float(y_str)
        radius = float(radius_str)
    except ValueError:
        raise ValidationError(
            "Locatie must be x: float, y: float, r: int"
        )

    # Check if the given coordinates are in RD. If they are not, convert them to WGS84.
    if y > 10:
        point = Point(x, y, srid=28992).transform(4326, clone=True)
    else:
        radius = dist_to_deg(radius, x)
        point = Point(x, y, srid=4326)

    return point, radius


def dist_to_deg(distance: float, latitude: float) -> float:
    """
    Convert meters to degrees.

    distance = distance in meters, latitude = latitude in degrees

    At the equator, the distance of one degree is equal in latitude and longitude.
    at higher latitudes, a degree longitude is shorter (in length), proportional to cos(latitude)
    http://en.wikipedia.org/wiki/Decimal_degrees
    This function is part of a distance filter where the database 'distance' is in degrees.
    There's no good single-valued answer to this problem.
    The distance/ degree is quite constant N/S around the earth (latitude),
    but varies over a huge range E/W (longitude).
    Split the difference: I'm going to average the degrees latitude and degrees longitude
    corresponding to the given distance. At high latitudes, this will be too short N/S
    and too long E/W. It splits the errors between the two axes.
    Errors are < 25 percent for latitudes < 60 degrees N/S.

    Args:
        distance: The distance in meters.
        latitude: The latitude in degrees.

    Returns:
        The distance in degrees.
    """
    #   d * (180 / pi) / earthRadius   ==> degrees longitude
    #   (degrees longitude) / cos(latitude)  ==> degrees latitude

    lat = latitude if latitude >= 0 else -1 * latitude
    rad2deg = 180 / pi
    earth_radius = 6378160.0
    latitude_correction = 0.5 * (1 + cos(lat * pi / 180))

    return distance / (earth_radius * latitude_correction) * rad2deg


def determine_bbox(request: Request) -> tuple[list[float] | None, str | None]:
    """
    Create a bounding box if it is given with the request.
    Returns the default bounding box if the bbox query parameter is found in the request.

    Args:
        request: A web request object.

    Returns:
        The bounding box and error message. The bounding box is a list of four coordinates
        [min_lon, min_lat, max_lon, max_lat]. The error message is a string or None.
    """
    if 'bbox' not in request.query_params:
        # Return the default bounding box if no bounding box is found in the request.
        return BBOX, None

    bboxp = request.query_params['bbox']
    bbox, err = valid_bbox(bboxp)

    if err:
        return None, err

    return bbox, err


def valid_bbox(bboxp: str, srid: int = 4326) -> tuple[list[float], str | None]: # noqa
    """
    Check if bbox is a valid bounding box. (wgs84) for now.

    Args:
        bboxp: A string containing the bounding box coordinates, in the format "min_lon,min_lat,max_lon,max_lat".
        srid: The spatial reference system of the bounding box.

    Returns:
        The bounding box and error message. The bounding box is a list of four coordinates
        [min_lon, min_lat, max_lon, max_lat]. The error message is a string or None.
    """
    bbox_str = bboxp.split(',')
    err = None

    # check if we got 4 parameters
    if len(bbox_str) != 4:
        return [], "wrong numer of arguments (lon, lat, lon, lat)"

    # check if we got floats
    try:
        bbox = [float(f) for f in bbox_str]
    except ValueError:
        return [], "Did not receive floats"

    # max bbox sizes from mapserver
    # RD  EXTENT      100000    450000   150000 500000
    # WGS             52.03560, 4.58565  52.48769, 5.31360
    lat_min = 52.03560
    lat_max = 52.48769
    lon_min = 4.58565
    lon_max = 5.31360

    if srid == 28992:
        # RD bbox from mapserver
        # 94000 465000 170000 514000
        lat_min = 465000
        lat_max = 514000
        lon_min = 94000
        lon_max = 170000

    # bbox given by leaflet
    lon1, lat1, lon2, lat2 = bbox

    if not lat_max >= lat1 >= lat_min:
        err = f"lat not within max bbox {lat_max} > {lat1} > {lat_min}"

    if not lat_max >= lat2 >= lat_min:
        err = f"lat not within max bbox {lat_max} > {lat2} > {lat_min}"

    if not lon_max >= lon2 >= lon_min:
        err = f"lon not within max bbox {lon_max} > {lon2} > {lon_min}"

    if not lon_max >= lon1 >= lon_min:
        err = f"lon not within max bbox {lon_max} > {lon1} > {lon_min}"

    bbox = [lon1, lat1, lon2, lat2]
    return bbox, err



"""
Serialization classes for Datapunt style Django REST Framework APIs.
"""
import json
from collections import OrderedDict
from typing import TYPE_CHECKING, Any, Generic, Mapping, TypedDict, TypeVar

from django.contrib.gis.geos import MultiPolygon, Point, Polygon
from django.db.models import Model
from django.http import HttpRequest
from rest_framework import serializers
from rest_framework.request import Request
from rest_framework.reverse import reverse

_IN = TypeVar("_IN")
_MT = TypeVar("_MT", bound=Model)


def get_links(
        view_name: str,
        kwargs: Mapping[str, Any] | None = None,
        request: HttpRequest | None = None
) -> OrderedDict[str, dict[str, str]]:
    result = OrderedDict([
        ('self', dict(
            href=reverse(view_name, kwargs=kwargs, request=request)
        ))
    ])

    return result


class DataSetSerializerMixin(serializers.BaseSerializer[_IN]):
    """Add dataset field to indicate 'source' of this data."""
    dataset: str

    def to_representation(self, obj: _IN) -> dict[str, Any]:
        result = super().to_representation(obj)

        assert isinstance(result, dict)

        result['dataset'] = self.dataset

        return result


if TYPE_CHECKING:
    class BaseLinksField(serializers.RelatedField[_MT, str, dict[str, dict[str, str | None]]]):
        pass
else:
    class BaseLinksField(Generic[_MT], serializers.RelatedField):
        pass


class LinksField(BaseLinksField[_MT]):
    lookup_field: str = 'pk'
    lookup_url_kwarg: str
    view_name: str | None = None

    def __init__(self, view_name: str | None = None, **kwargs: Any):
        if view_name is not None:
            self.view_name = view_name
        assert self.view_name is not None, 'The `view_name` argument is required.'
        self.lookup_field = kwargs.pop('lookup_field', self.lookup_field)
        self.lookup_url_kwarg = kwargs.pop('lookup_url_kwarg', self.lookup_field)

        kwargs['read_only'] = True
        kwargs['source'] = '*'

        super().__init__(**kwargs)

    def get_url(self, obj: Model, view_name: str, request: Request | None, format: str | None) -> str | None:
        """
        Given an object, return the URL that hyperlinks to the object.

        May raise a `NoReverseMatch` if the `view_name` and `lookup_field`
        attributes are not configured to correctly match the URL conf.
        """
        # Unsaved objects will not yet have a valid URL.
        if hasattr(obj, 'pk') and obj.pk in (None, ''):
            return None

        lookup_value = getattr(obj, self.lookup_field)
        kwargs = {self.lookup_url_kwarg: lookup_value}

        return reverse(view_name, kwargs=kwargs, request=request, format=format)

    def to_representation(self, value: _MT) -> dict[str, dict[str, str | None]]:
        request = self.context.get('request')
        assert self.view_name is not None

        return OrderedDict([('self', {
            'href': self.get_url(value, self.view_name, request, None)
        })])


class HALSerializer(serializers.HyperlinkedModelSerializer):
    url_field_name: str = '_links'
    serializer_url_field = LinksField


class SelfLinkSerializerMixin(serializers.BaseSerializer[_IN]):
    def get__links(self, obj: Model) -> dict[str, dict[str, str]]:
        """
        Serialization of _links field for detail view (assumes ModelViewSet).

        Note:
            Used to provide HAL-JSON style self links.
        """
        view = self.context['view']
        model = view.queryset.model
        pk_value = getattr(obj, model._meta.pk.name)

        return {
            'self': {
                'href': view.reverse_action('detail', args=[pk_value])
            }
        }


class Summary(TypedDict):
    count: int
    href: str


class RelatedSummaryField(serializers.Field):
    def to_representation(self, value) -> Summary:
        count = value.count()
        model_name = value.model.__name__
        mapping = model_name.lower() + "-list"
        url = reverse(mapping, request=self.context['request'])

        parent_pk = value.instance.pk
        filter_name = list(value.core_filters.keys())[0]

        return {
            'count': count,
            'href': f'{url}?{filter_name}={parent_pk}',
        }


# Note about DisplayField below; setting source to '*' causes the
# whole (model) instance to be passed to the DisplayField See:
# http://www.django-rest-framework.org/api-guide/fields/#source
# Display field then uses the __str__ function on the Django
# model to get a nice string representation that can be presented
# to the user.

if TYPE_CHECKING:
    class BaseDisplayField(serializers.Field[_MT, str, str, Any]):
        pass
else:
    class BaseDisplayField(Generic[_MT], serializers.Field):
        pass


class DisplayField(BaseDisplayField[_MT]):
    """
    Add a `_display` field, based on Model string representation.
    """
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        kwargs['source'] = '*'
        kwargs['read_only'] = True
        super().__init__(*args, **kwargs)

    def to_representation(self, value: _MT) -> str:
        return str(value)


class GeoJson(TypedDict):
    type: str
    coordinates: list[float] | list[list[list[float]]] | list[list[list[list[float]]]]


class BaseMultipleGeometry:
    """Extend this class when you plan to use the serializer field below."""
    geometrie: Point | Polygon | MultiPolygon | None


if TYPE_CHECKING:
    BaseMultipleGeometryField =\
        serializers.Field[Point | Polygon | MultiPolygon | None, GeoJson, str | GeoJson, BaseMultipleGeometry]
else:
    BaseMultipleGeometryField = serializers.Field


class MultipleGeometryField(BaseMultipleGeometryField):
    read_only: bool = True

    def to_representation(self, value: Point | Polygon | MultiPolygon | None) -> str | GeoJson:
        res = ''
        if value:
            res = json.loads(value.geojson)

        return res



from django.test import TestCase
from rest_framework.test import APIClient

from datapunt_api.pagination import HALCursorPagination
from tests.models import Person
from tests.views import PersonViewSet


class HALPaginationTest(TestCase):
    def setUp(self):
        for x in range(10):
            person = Person()
            person.name = f'person #{x}'
            person.save()

    def test_pagination_one_page(self):
        client = APIClient()
        response = client.get('/tests/person/?page_size=10', format='json')

        self.assertEqual(response.status_code, 200)
        payload = response.json()

        self.assertIsNone(payload['_links']['previous']['href'])
        self.assertIsNone(payload['_links']['next']['href'])

        self.assertEqual(payload['count'], 10)
        self.assertEqual(len(payload['results']), 10)

    def test_pagination_multiple_pages(self):
        client = APIClient()
        response = client.get('/tests/person/?page_size=2&page=2', format='json')

        self.assertEqual(response.status_code, 200)
        payload = response.json()

        self.assertIsNotNone(payload['_links']['previous']['href'])
        self.assertEqual(payload['_links']['previous']['href'],
                         'http://testserver/tests/person/?page=1&page_size=2')

        self.assertIsNotNone(payload['_links']['next']['href'])
        self.assertEqual(payload['_links']['next']['href'],
                         'http://testserver/tests/person/?page=3&page_size=2')

        self.assertEqual(payload['count'], 10)
        self.assertEqual(len(payload['results']), 2)

    def test_pagination_url_endswith_api(self):
        client = APIClient()
        response = client.get('/tests/person.api', format='json')

        self.assertEqual(response.status_code, 200)
        self.assertNotEqual(response.data['_links']['self']['href'], 'http://testserver/tests/person.api')
        self.assertEqual(response.data['_links']['self']['href'], 'http://testserver/tests/person')


class HALCursorPaginationTest(TestCase):
    def setUp(self):
        setattr(PersonViewSet, 'pagination_class', HALCursorPagination)
        setattr(HALCursorPagination, 'ordering', ['id'])

        for x in range(10):
            person = Person()
            person.name = f'person #{x}'
            person.save()

    def test_pagination_one_page(self):
        client = APIClient()
        response = client.get('/tests/person/?page_size=10', format='json')

        self.assertEqual(response.status_code, 200)
        payload = response.json()

        self.assertIsNone(payload['_links']['previous']['href'])
        self.assertIsNone(payload['_links']['next']['href'])

        self.assertEqual(payload['count'], 10)
        self.assertEqual(len(payload['results']), 10)

    def test_pagination_multiple_pages(self):
        client = APIClient()
        response = client.get('/tests/person/?page_size=2', format='json')

        self.assertEqual(response.status_code, 200)
        payload = response.json()

        self.assertIsNone(payload['_links']['previous']['href'])
        self.assertIsNotNone(payload['_links']['next']['href'])
        self.assertIn('cursor=', payload['_links']['next']['href'])

        self.assertEqual(payload['count'], 10)
        self.assertEqual(len(payload['results']), 2)

    def test_pagination_without_count(self):
        setattr(HALCursorPagination, 'count_table', False)

        client = APIClient()
        response = client.get('/tests/person/?page_size=10', format='json')

        self.assertEqual(response.status_code, 200)
        payload = response.json()

        self.assertNotIn('count', payload)
        self.assertEqual(len(payload['results']), 10)



from typing import TYPE_CHECKING, Any, Generic, TypeVar

from django.db.models import Model, QuerySet
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import renderers, viewsets
from rest_framework.pagination import BasePagination
from rest_framework.renderers import BaseRenderer
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.settings import api_settings
from rest_framework.views import APIView
from rest_framework_extensions.mixins import DetailSerializerMixin
from rest_framework_xml.renderers import XMLRenderer

from .pagination import HALPagination
from .renderers import PaginatedCSVRenderer
from .serializers import (DisplayField, HALSerializer,  # noqa: F401
                          LinksField, RelatedSummaryField,
                          SelfLinkSerializerMixin, get_links)

_MT = TypeVar("_MT", bound=Model)
_MT_co = TypeVar("_MT_co", bound=Model, covariant=True)


DEFAULT_RENDERERS: list[type[BaseRenderer]] = [
    renderers.JSONRenderer,
    PaginatedCSVRenderer,
    renderers.BrowsableAPIRenderer,
    XMLRenderer,
]


if api_settings.DEFAULT_RENDERER_CLASSES:
    # The APISettings class automagically converts the strings to classes
    DEFAULT_RENDERERS = api_settings.DEFAULT_RENDERER_CLASSES  # type: ignore


class _DisabledHTMLFilterBackend(DjangoFilterBackend):
    """See https://github.com/tomchristie/django-rest-framework/issues/3766.

    This prevents DRF from generating the filter dropdowns (which can be HUGE
    in our case)
    """

    def to_html(self, request: Request, queryset: QuerySet[_MT], view: APIView) -> str:
        return ""


def _is_detailed_request(detailed_keyword: str, request: Request) -> bool:
    value = request.GET.get(detailed_keyword, False)
    if value and value in [1, '1', True, 'True', 'yes']:
        return True

    return False


if TYPE_CHECKING:
    class ReadOnlyModelViewSet(viewsets.ReadOnlyModelViewSet[_MT_co]):
        pass

    class BaseDetailSerializerMixin(DetailSerializerMixin[_MT_co]):
        pass
else:
    class ReadOnlyModelViewSet(Generic[_MT_co], viewsets.ReadOnlyModelViewSet):
        pass

    class BaseDetailSerializerMixin(Generic[_MT_co], DetailSerializerMixin):
        pass


class DatapuntViewSet(BaseDetailSerializerMixin[_MT_co], ReadOnlyModelViewSet[_MT_co]):
    """ViewSet subclass for use in Datapunt APIs.

    Note:
    - this uses HAL JSON style pagination.
    """

    renderer_classes: list[type[BaseRenderer]] = DEFAULT_RENDERERS
    pagination_class: type[BasePagination] = HALPagination
    filter_backends = [_DisabledHTMLFilterBackend]
    # TO restore filter box in your view!
    # filter_backends = [DjangoFilterBackend]
    detailed_keyword: str = 'detailed'

    def list(self, request: Request, *args: Any, **kwargs: Any) -> Response:
        # Checking if a detailed response is required
        if _is_detailed_request(self.detailed_keyword, request):
            self.serializer_class = self.serializer_detail_class
        return super().list(request, *args, **kwargs)

    def get_renderer_context(self) -> dict[str, Any]:
        # make CSV select fields to render.
        context = super().get_renderer_context()
        context['header'] = (
            self.request.GET['fields'].split(',')
            if 'fields' in self.request.GET else None)
        return context


if TYPE_CHECKING:
    class ModelViewSet(viewsets.ModelViewSet[_MT_co]):
        pass
else:
    class ModelViewSet(Generic[_MT_co], viewsets.ModelViewSet):
        pass


class DatapuntViewSetWritable(BaseDetailSerializerMixin[_MT_co], ModelViewSet[_MT_co]):
    """ViewSet subclass for use in Datapunt APIs.

    Note:
    - this uses HAL JSON style pagination.
    """

    renderer_classes: list[type[BaseRenderer]] = DEFAULT_RENDERERS
    pagination_class: type[BasePagination] = HALPagination
    filter_backends = [_DisabledHTMLFilterBackend]

    detailed_keyword: str = 'detailed'

    def list(self, request: Request, *args: Any, **kwargs: Any) -> Response:
        # Checking if a detailed response is required
        if _is_detailed_request(self.detailed_keyword, request):
            self.serializer_class = self.serializer_detail_class
        return super().list(request, *args, **kwargs)



from rest_framework import routers

from .views import (PersonViewSet, SimpleViewSet, TemperatureRecordViewSet,
                    ThingViewSet, WeatherStationViewSet)


class ClimateAPI(routers.APIRootView):
    """
    Example API KNMI climate data.

    Note: this is used for testing drf_amsterdam project.
    """

    def get_api_root_view(self, **kwargs):
        view = super().get_api_root_view(**kwargs)
        cls = view.cls

        class Climate(cls):
            def get_view_name(self):
                return 'Example API KNMI climate data'

        Climate.__doc__ = self.__doc__
        return Climate.as_view()


class ApiRouter(routers.DefaultRouter):
    """The main router"""
    APIRootView = ClimateAPI


router = ApiRouter()
router.register('weatherstation', WeatherStationViewSet)
router.register('temperature_record', TemperatureRecordViewSet)
router.register('simple', SimpleViewSet)
router.register('thing', ThingViewSet)
router.register('person', PersonViewSet)
router.register('person.api', PersonViewSet)

urlpatterns = router.urls



import os

from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='drf_amsterdam',
    version='1.0.0',
    packages=find_packages(),
    install_requires=[
        'djangorestframework',
        'drf_extensions',
        'djangorestframework_csv',
        'djangorestframework_xml'],
    include_package_data=True,
    package_data={'datapunt_api': ['py.typed']},
    license='Mozilla Public License Version 2.0',
    description='Amsterdam Datapunt code and style for Django REST Framework.',
    long_description=README,
    author='Amsterdam Datapunt',
    author_email='datapunt@amsterdam.nl',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.11',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)'
    ],
)



"""Test views."""
from django_filters.rest_framework import (DjangoFilterBackend, FilterSet,
                                           filters)
from rest_framework.filters import OrderingFilter
from rest_framework.viewsets import ReadOnlyModelViewSet

from datapunt_api import bbox
from datapunt_api.rest import (DatapuntViewSet, DatapuntViewSetWritable,
                               _DisabledHTMLFilterBackend)
from tests.models import (Person, SimpleModel, TemperatureRecord, Thing,
                          WeatherStation)
from tests.serializers import (DetailedPersonSerializer, PersonSerializer,
                               SelfLinksSerializer,
                               TemperatureRecordSerializer, ThingSerializer,
                               WeatherDetailStationSerializer,
                               WeatherStationSerializer)


class WeatherFilter(FilterSet):
    """Weather station fitler.

    test geo location filtering
    """

    location = filters.CharFilter(
        method="locatie_filter", label='x,y,r')

    location_rd = filters.CharFilter(
        method="locatie_filter_rd", label='x,y,r')

    def locatie_filter(self, qs, name, value):  # noqa
        point, radius = bbox.parse_xyr(value)
        return qs.filter(
            centroid__geometrie__dwithin=(point, radius))

    def locatie_filter_rd(self, qs, name, value):  # noqa
        point, radius = bbox.parse_xyr(value, srid=28992)
        return qs.filter(
            centroid_rd__dwithin=(point, radius))

    class Meta(object): # noqa
        model = WeatherStation

        fields = (
            "location",
            "location_rd",
        )


class WeatherStationViewSet(DatapuntViewSet):  # noqa
    serializer_class = WeatherStationSerializer
    serializer_detail_class = WeatherDetailStationSerializer
    queryset = WeatherStation.objects.all().order_by('id')

    filter_class = WeatherFilter

    filter_backends = (
        _DisabledHTMLFilterBackend,
        DjangoFilterBackend,
        OrderingFilter
    )

    ordering_fields = '__all__'


class TemperatureRecordViewSet(DatapuntViewSet): # noqa
    serializer_class = TemperatureRecordSerializer
    serializer_detail_class = TemperatureRecordSerializer
    queryset = TemperatureRecord.objects.all().order_by('date')

    # test custom inheritance.
    def list(self, request, *args, **kwargs):  # noqa
        return super().list(self, request, *args, **kwargs)


class SimpleViewSet(ReadOnlyModelViewSet):
    queryset = SimpleModel.objects.all().order_by('id')
    serializer_class = SelfLinksSerializer


class ThingViewSet(ReadOnlyModelViewSet):
    queryset = Thing.objects.all().order_by('id')
    serializer_class = ThingSerializer


class PersonViewSet(DatapuntViewSetWritable):
    queryset = Person.objects.all().order_by('id')
    serializer_class = PersonSerializer
    serializer_detail_class = DetailedPersonSerializer



"""Test models."""
from django.contrib.gis.db import models
from django.contrib.gis.db.models import PointField


class WeatherStation(models.Model):  # noqa
    number = models.IntegerField(unique=True)

    centroid = models.PointField(name='centroid', srid=4326)
    centroid_rd = models.PointField(name='centroid_rd', srid=28992)

    def __str__(self):  # noqa
        return 'DISPLAY FIELD CONTENT'


class TemperatureRecord(models.Model):  # noqa
    class Meta:  # noqa
        unique_together = ('station', 'date')

    station = models.ForeignKey(WeatherStation, on_delete=models.CASCADE)
    date = models.DateField()
    temperature = models.DecimalField(decimal_places=3, max_digits=6)


class SimpleModel(models.Model):
    name = models.CharField(max_length=25)
    age = models.IntegerField()
    sign = models.CharField(max_length=25)


class Person(models.Model):
    name = models.CharField(max_length=255)


class Thing(models.Model):
    name = models.CharField(max_length=128)
    person = models.ForeignKey(Person, related_name='things', on_delete=models.DO_NOTHING)


class Location(models.Model):
    geometrie = PointField(srid=28992)
